@@ -1,7 +1,7 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 /*
  * libthai - Thai Language Support Library
- * Copyright (C) 2001  Theppitak Karoonboonyanan <thep@linux.thai.net>
+ * Copyright (C) 2001  Theppitak Karoonboonyanan <theppitak@gmail.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -21,7 +21,7 @@
 /*
  * brk-maximal.c - Thai word break with maximal matching scheme
  * Created: 2006-08-20
- * Author:  Theppitak Karoonboonyanan <thep@linux.thai.net>
+ * Author:  Theppitak Karoonboonyanan <theppitak@gmail.com>
  */
 
 #include <string.h>
@@ -31,6 +31,7 @@
 #include <datrie/trie.h>
 #include <thai/tis.h>
 #include <thai/thwchar.h>
+#include "thbrk-utils.h"
 #include "brk-maximal.h"
 #include "brk-common.h"
 
@@ -101,13 +102,13 @@ typedef struct {
  *----------------------------------*/
 
 static BrkPool *    brk_root_pool (int pos_size);
-static int          brk_maximal_do_impl (const thchar_t *s, int len,
+static int          brk_maximal_do_impl (const thwchar_t *ws, int len,
                                          const char *brkpos_hints,
                                          int pos[], size_t n);
-static int          brk_recover_try (const thchar_t *s, int len,
+static int          brk_recover_try (const thwchar_t *ws, int len,
                                      const char *brkpos_hints,
                                      size_t recov_words, int *last_brk_pos);
-static int          brk_recover (const thchar_t *text, int len, int pos,
+static int          brk_recover (const thwchar_t *wtext, int len, int pos,
                                  const char *brkpos_hints, RecovHist *rh);
 
 /*---------------------*
@@ -129,20 +130,34 @@ int
 brk_maximal_do (const thchar_t *s, int len, int pos[], size_t n)
 {
     char        *brkpos_hints;
+    thwchar_t   *ws;
     int          ret;
 
     brkpos_hints = (char *) malloc (len);
+    if (UNLIKELY (!brkpos_hints))
+        goto err_nothing_done;
     brk_brkpos_hints (s, len, brkpos_hints);
 
-    ret = brk_maximal_do_impl (s, len, brkpos_hints, pos, n);
+    ws = (thwchar_t *) malloc ((len + 1) * sizeof (thwchar_t));
+    if (UNLIKELY (!ws))
+        goto err_brkpos_hints_created;
+    th_tis2uni_line (s, ws, len + 1);
 
+    ret = brk_maximal_do_impl (ws, len, brkpos_hints, pos, n);
+
+    free (ws);
     free (brkpos_hints);
 
     return ret;
+
+err_brkpos_hints_created:
+    free (brkpos_hints);
+err_nothing_done:
+    return 0;
 }
 
 static int
-brk_maximal_do_impl (const thchar_t *s, int len,
+brk_maximal_do_impl (const thwchar_t *ws, int len,
                      const char *brkpos_hints,
                      int pos[], size_t n)
 {
@@ -154,7 +169,7 @@ brk_maximal_do_impl (const thchar_t *s, int len,
 
     pool = brk_root_pool (n);
     best_brk = best_brk_new (n);
-    if (!best_brk)
+    if (UNLIKELY (!best_brk))
         return 0;
     recov_hist.pos = recov_hist.recov = -1;
 
@@ -169,14 +184,14 @@ brk_maximal_do_impl (const thchar_t *s, int len,
         is_recovered = 0;
         str_pos = shot->str_pos;
         do {
-            if (!trie_state_walk (shot->dict_state, th_tis2uni (s[str_pos++])))
+            if (!trie_state_walk (shot->dict_state, ws[str_pos++]))
             {
                 int recovered;
 
                 is_terminal = 0;
 
                 /* try to recover from error */
-                recovered = brk_recover (s, len, shot->str_pos + 1,
+                recovered = brk_recover (ws, len, shot->str_pos + 1,
                                          brkpos_hints, &recov_hist);
                 if (-1 != recovered) {
                     /* add penalty by recovered - recent break pos */
@@ -262,7 +277,7 @@ brk_maximal_do_impl (const thchar_t *s, int len,
 }
 
 static int
-brk_recover_try (const thchar_t *s, int len,
+brk_recover_try (const thwchar_t *ws, int len,
                  const char *brkpos_hints,
                  size_t recov_words, int *last_brk_pos)
 {
@@ -282,8 +297,7 @@ brk_recover_try (const thchar_t *s, int len,
         is_keep_node = 1;
         for (;;) {
             do {
-                if (!trie_state_walk (shot->dict_state,
-                                      th_tis2uni (s[shot->str_pos++])))
+                if (!trie_state_walk (shot->dict_state, ws[shot->str_pos++]))
                 {
                     is_keep_node = 0;
                     break;
@@ -351,7 +365,8 @@ brk_root_pool (int pos_size)
 
     pool = NULL;
 
-    if (NULL == (dict = brk_get_dict()))
+    dict = brk_get_dict();
+    if (UNLIKELY (!dict))
         return NULL;
     root_shot.dict_state = trie_root (dict);
     root_shot.brk_pos = NULL; /* it's not used anyway */
@@ -360,7 +375,7 @@ brk_root_pool (int pos_size)
     root_shot.penalty = 0;
 
     node = brk_pool_node_new (&root_shot);
-    if (node) {
+    if (LIKELY (node)) {
         pool = brk_pool_add (pool, node);
     }
 
@@ -372,24 +387,24 @@ brk_root_pool (int pos_size)
 #define RECOVERED_WORDS 3
 
 static int
-brk_recover (const thchar_t *text, int len, int pos,
+brk_recover (const thwchar_t *wtext, int len, int pos,
              const char *brkpos_hints, RecovHist *rh)
 {
-    int last_brk_pos;
+    int last_brk_pos = 0;
     int n, p;
 
     while (pos < len && !brkpos_hints[pos]) {
         ++pos;
     }
-    if (rh->pos == pos)
+    if (UNLIKELY (rh->pos == pos))
         return rh->recov;
 
     for (p = pos; p < len; ++p) {
         if (brkpos_hints[p]) {
-            n = brk_recover_try (text + p, len - p, brkpos_hints + p,
+            n = brk_recover_try (wtext + p, len - p, brkpos_hints + p,
                                  RECOVERED_WORDS, &last_brk_pos);
             if (n == RECOVERED_WORDS
-                || (n > 0 && '\0' == text[last_brk_pos]))
+                || (n > 0 && '\0' == wtext[last_brk_pos]))
             {
                 rh->pos = pos;
                 rh->recov = p;
@@ -407,7 +422,7 @@ brk_shot_init (BrkShot *dst, const BrkShot *src)
     dst->dict_state = trie_state_clone (src->dict_state);
     dst->str_pos = src->str_pos;
     dst->brk_pos = (int *) malloc (src->n_brk_pos * sizeof (int));
-    if (!dst->brk_pos)
+    if (UNLIKELY (!dst->brk_pos))
         return -1;
     memcpy (dst->brk_pos, src->brk_pos, src->cur_brk_pos * sizeof (int));
     dst->n_brk_pos = src->n_brk_pos;
@@ -478,9 +493,9 @@ brk_pool_node_new (const BrkShot *shot)
         brk_shot_reuse (&node->shot, shot);
     } else {
         node = (BrkPool *) malloc (sizeof (BrkPool));
-        if (!node)
+        if (UNLIKELY (!node))
             return NULL;
-        if (brk_shot_init (&node->shot, shot) != 0) {
+        if (UNLIKELY (brk_shot_init (&node->shot, shot) != 0)) {
             free (node);
             return NULL;
         }
@@ -592,15 +607,15 @@ best_brk_new (int n_brk_pos)
 {
     BestBrk *best_brk;
 
-    if ((size_t) n_brk_pos > SIZE_MAX / sizeof (int))
+    if (UNLIKELY ((size_t) n_brk_pos > SIZE_MAX / sizeof (int)))
         return NULL;
 
     best_brk = (BestBrk *) malloc (sizeof (BestBrk));
-    if (!best_brk)
+    if (UNLIKELY (!best_brk))
         return NULL;
 
     best_brk->brk_pos = (int *) malloc ((size_t) n_brk_pos * sizeof (int));
-    if (!best_brk->brk_pos)
+    if (UNLIKELY (!best_brk->brk_pos))
         goto exit1;
     best_brk->n_brk_pos = n_brk_pos;
     best_brk->cur_brk_pos = best_brk->str_pos = 0;
